/*
 * Copyright 2022 The TensorFlow Runtime Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <utility>

#include "mlir/Dialect/Bufferization/Transforms/Bufferize.h"
#include "tensorflow/compiler/jit/flags.h"
#include "tensorflow/compiler/mlir/tensorflow/dialect_registration.h"
#include "tensorflow/compiler/mlir/tfrt/jit/tf_jitrt_pipeline.h"
#include "tensorflow/compiler/xla/mlir/runtime/transforms/compiler.h"
#include "tensorflow/compiler/xla/runtime/executable.h"
#include "tensorflow/compiler/xla/runtime/jit_executable.h"
#include "tensorflow/core/platform/test_benchmark.h"
#include "tfrt/jitrt/jitrt_compiler.h"  // from @tf_runtime

namespace tensorflow {

static const char* mlir_module = R"(
  func.func @compute(
      %arg0: tensor<?x?x1xf32>,
      %arg1: tensor<?x?x20xf32>,
      %arg2: tensor<?x?x40xf32>,
      %arg3: tensor<?x?x40xf32>,
      %arg4: tensor<?x?x33xf32>,
      %arg5: tensor<?x?x13xf32>
  ) -> (tensor<?x?x20xf32>, tensor<?x?x40xf32>, tensor<?x?x40xf32>,
        tensor<?x?x33xf32>, tensor<?x?x13xf32>)
  {
    %c = "tf.Const"() {value = dense<9.99999997E-7> : tensor<f32>}
         : () -> tensor<f32>
    %0 = "tf.AddV2"(%arg0, %c)
         : (tensor<?x?x1xf32>, tensor<f32>) -> tensor<?x?x1xf32>
    %1 = "tf.RealDiv"(%arg1, %0)
         : (tensor<?x?x20xf32>, tensor<?x?x1xf32>) -> tensor<?x?x20xf32>
    %2 = "tf.RealDiv"(%arg2, %0)
         : (tensor<?x?x40xf32>, tensor<?x?x1xf32>) -> tensor<?x?x40xf32>
    %3 = "tf.RealDiv"(%arg3, %0)
         : (tensor<?x?x40xf32>, tensor<?x?x1xf32>) -> tensor<?x?x40xf32>
    %4 = "tf.RealDiv"(%arg4, %0)
         : (tensor<?x?x33xf32>, tensor<?x?x1xf32>) -> tensor<?x?x33xf32>
    %5 = "tf.RealDiv"(%arg5, %0)
         : (tensor<?x?x13xf32>, tensor<?x?x1xf32>) -> tensor<?x?x13xf32>
    return %1, %2, %3, %4, %5 : tensor<?x?x20xf32>, tensor<?x?x40xf32>,
                                tensor<?x?x40xf32>, tensor<?x?x33xf32>,
                                tensor<?x?x13xf32>
  })";

static const char* entrypoint = "compute";

using ::tfrt::jitrt::CompilationPipelineOptions;
using ::tfrt::jitrt::CreateDefaultJitRtCompilationPipeline;
using ::tfrt::jitrt::RegisterDefaultJitRtDialects;

using ::xla::runtime::AsyncValuesCache;
using ::xla::runtime::JitExecutable;

using JitExecutableCache = AsyncValuesCache<size_t, JitExecutable>;

static void BM_InstantiateExecutable(::testing::benchmark::State& state) {
  // Options for the default JitRt compilation pipeline (lowering to LLVM).
  CompilationPipelineOptions copts;
  copts.alignment = EIGEN_MAX_ALIGN_BYTES;
  copts.num_worker_threads = 16;
  copts.cost_driven_async_parallel_for = false;

  // Options for the JitRt JitExecutable compilation.
  JitExecutable::Options opts;
  opts.specialization = JitExecutable::Specialization::kEnabled;

  // Register dialects and interfaces required for the compilation pipeline.
  opts.compiler.register_dialects =
      [](xla::runtime::DialectRegistry& dialects) {
        mlir::RegisterAllTensorFlowDialects(*dialects);
        RegisterDefaultJitRtDialects(dialects);
      };

  // Register a custom pipeline for lowering from Tensorflow dialect to LLVM.
  opts.compiler.create_compilation_pipeline =
      [&](xla::runtime::PassManager& passes) {
        TfJitRtPipelineOptions opts;
        opts.enable_xla_cpu_transformations =
            tensorflow::GetJitRtFlags().enable_xla_cpu_transformations;
        opts.lower_to_mmt4d = tensorflow::GetJitRtFlags().pack_matmul;

        // Lower from Tensorflow to Linalg on buffers.
        CreateTfJitRtPipeline(*passes, opts);

        // Use default JitRt compilation pipeline to lower to LLVM.
        CreateDefaultJitRtCompilationPipeline(passes, copts);
      };

  // Register a custom pipeline to propagate specialization information.
  opts.compiler.create_specialization_pipeline =
      [&](xla::runtime::PassManager& passes) {
        CreateJitRtSpecializationPipeline(passes);
      };

  // When lowering Tensorflow functions to JitRt we convert all input and
  // result tensors to memrefs, and add a kernel context input.
  opts.compiler.calling_convention = xla::runtime::DefaultCallingConvention(
      mlir::bufferization::BufferizeTypeConverter());

  for (auto _ : state) {
    absl::StatusOr<JitExecutable> jit_executable =
        JitExecutable::Instantiate(mlir_module, entrypoint, opts, "benchmark");
    if (!jit_executable.ok()) {
      LOG(FATAL) << "Failed to compile the kernel: "
                 << jit_executable.status().message();
    }
  }
}

BENCHMARK(BM_InstantiateExecutable);

}  // namespace tensorflow
